"""Base class for models."""

import abc

import tensorflow as tf

from opennmt import optimizers
from opennmt import schedules
from opennmt.utils import exporters
from opennmt.utils import losses
from opennmt.utils import misc


class Model(tf.keras.layers.Layer):
  """Base class for models."""

  def __init__(self, examples_inputter):
    super(Model, self).__init__()
    self.examples_inputter = examples_inputter
    self.params = {}

  @property
  def unsupervised(self):
    """Unsupervised model."""
    return self.labels_inputter is None

  @property
  def features_inputter(self):
    """The inputter producing features."""
    return getattr(self.examples_inputter, "features_inputter", self.examples_inputter)

  @property
  def labels_inputter(self):
    """The inputter producing labels."""
    return getattr(self.examples_inputter, "labels_inputter", None)

  @property
  def ctranslate2_spec(self):
    """The equivalent CTranslate2 model specification."""
    return None

  def auto_config(self, num_replicas=1):
    """Returns automatic configuration values specific to this model.

    Args:
      num_replicas: The number of synchronous model replicas used for the
        training.

    Returns:
      A partial training configuration.
    """
    _ = num_replicas
    return {}

  def initialize(self, data_config, params=None):
    """Initializes the model from the data configuration.

    Args:
      data_config: A dictionary containing the data configuration set
        by the user (e.g. vocabularies, tokenization, pretrained embeddings,
        etc.).
      params: A dictionary of hyperparameters.
    """
    if params is None:
      params = {}
    self.params.update(params)
    dropout = self.params.get("dropout")
    if dropout is not None:
      misc.set_dropout(self, dropout)
    self.examples_inputter.initialize(data_config)

  def build(self, input_shape):
    freeze_layers = self.params.get("freeze_layers")
    if freeze_layers:
      if not isinstance(freeze_layers, list):
        freeze_layers = [freeze_layers]
      for layer_path in freeze_layers:
        layer = misc.index_structure(self, layer_path)
        layer.trainable = False
        misc.set_dropout(layer, 0)  # Disable dropout in frozen layers.
    self.examples_inputter.build(input_shape)
    self.built = True

  @abc.abstractmethod
  def call(self, features, labels=None, training=None, step=None):  # pylint: disable=arguments-differ
    """Runs the model.

    Args:
      features: A nested structure of features ``tf.Tensor``.
      labels: A nested structure of labels ``tf.Tensor``.
      training: If ``True``, run in training mode.
      step: The current training step.

    Returns:
      A tuple containing,

      - The model outputs (usually unscaled probabilities).
      - The model predictions.
    """
    raise NotImplementedError()

  def infer(self, features):
    """Runs inference on :obj:`features`.

    This is a small convenience wrapper around
    :meth:`opennmt.models.Model.call`.

    Args:
      features: A nested structure of features ``tf.Tensor``.

    Returns:
      The model predictions.
    """
    _, predictions = self(features)
    if "index" in features:
      predictions["index"] = features["index"]
    return predictions

  def evaluate(self, features, labels):
    """Evaluates :obj:`features` predictions against `labels`.

    Args:
      features: A nested structure of features ``tf.Tensor``.
      labels: A nested structure of labels ``tf.Tensor``.

    Returns:
      A tuple with the loss and the model predictions.
    """
    outputs, predictions = self(features, labels=labels)
    loss = self.compute_loss(outputs, labels, training=False)
    return loss, predictions

  def score(self, features, labels):
    """Scores labels.

    Args:
      features: A nested structure of features ``tf.Tensor``.
      labels: A nested structure of labels ``tf.Tensor``.

    Returns:
      The score results.
    """
    raise NotImplementedError("This model does not define a score function")

  @abc.abstractmethod
  def compute_loss(self, outputs, labels, training=True):
    """Computes the loss.

    Args:
      outputs: The model outputs (usually unscaled probabilities).
      labels: The dict of labels ``tf.Tensor``.
      training: If ``True``, compute the loss for training.

    Returns:
      The loss or a tuple ``(numerator, train_denominator, stats_denominator)``
      to use a different normalization for training compared to reporting (e.g.
      batch-normalized for training vs. token-normalized for reporting).
    """
    raise NotImplementedError()

  def regularize_loss(self, loss, variables=None):
    """Regularizes the loss.

    Args:
      loss: The loss.
      variables: List of variables.

    Returns:
      The regularized loss.
    """
    if variables is None:
      variables = self.trainable_variables
    regularization = self.params.get("regularization")
    if regularization is not None:
      loss += losses.regularization_penalty(
          regularization["type"], regularization["scale"], variables)
    return loss

  def get_metrics(self):
    """Returns the metrics for this model.

    Returns:
      A dictionary of ``tf.keras.metrics.Metric`` metrics.
    """
    return None

  def update_metrics(self, metrics, predictions, labels):  # pylint: disable=unused-argument
    """Computes additional metrics on the predictions.

    Args:
      metrics: A dictionary of metrics to update.
      predictions: The model predictions.
      labels: The dict of labels ``tf.Tensor``.
    """
    return

  def get_optimizer(self):
    """Returns the optimizer for this model.

    Returns:
      A ``tf.keras.optimizers.Optimizer`` instance or ``None`` if no optimizer
      is configured.
    """
    params = self.params
    optimizer_name = params.get("optimizer")
    if optimizer_name is None:
      return None
    learning_rate = tf.constant(params["learning_rate"], dtype=tf.float32)
    if params.get("decay_type") is not None:
      schedule_params = params.get("decay_params", {})
      learning_rate = schedules.make_learning_rate_schedule(
          learning_rate,
          params["decay_type"],
          schedule_params=schedule_params,
          start_step=params.get("start_decay_steps", 0),
          minimum_learning_rate=params.get("minimum_learning_rate", 0))
    optimizer_params = params.get("optimizer_params")
    if optimizer_params is None:
      optimizer_params = {}
    optimizer = optimizers.make_optimizer(
        optimizer_name, learning_rate, **optimizer_params)
    return optimizer

  def serve_function(self):
    """Returns a function for serving this model.

    Returns:
      A ``tf.function``.
    """
    # Set name attribute of the input TensorSpec.
    input_signature = {
        name:tf.TensorSpec.from_spec(spec, name=name)
        for name, spec in self.features_inputter.input_signature().items()}

    @tf.function(input_signature=(input_signature,))
    def _run(features):
      features = self.features_inputter.make_features(features=features.copy())
      if isinstance(features, (list, tuple)):
        # Special case for unsupervised inputters that always return a tuple (features, labels).
        features = features[0]
      _, predictions = self(features)
      return predictions

    return _run

  def export(self, export_dir, exporter=None):
    """Exports the model for serving.

    Args:
      export_dir: The output directory.
      exporter: A :class:`opennmt.utils.Exporter` instance. Defaults to
        :class:`opennmt.utils.SavedModelExporter`.
    """
    if exporter is None:
      exporter = exporters.SavedModelExporter()
    exporter.export(self, export_dir)

  def create_variables(self, optimizer=None):
    """Creates the model variables by running it once.

    Args:
      optimizer: If set, also create the optimizer variables.
    """
    # Create input features from the input signatures. We remove the leading
    # batch dimension as sometimes assumed by make_features methods and set
    # unspecified dimensions to 1.
    features = tf.nest.map_structure(
        lambda spec: tf.fill(
            [dim or 1 for dim in spec.shape.as_list()[1:]],
            tf.constant("a" if spec.dtype is tf.string else 1, dtype=spec.dtype)),
        self.examples_inputter.input_signature())
    features = self.examples_inputter.make_features(features=features)

    # Add the batch dimension back before calling the model.
    features, labels = tf.nest.map_structure(lambda x: tf.expand_dims(x, 0), features)
    _ = self(features, labels=labels, training=True, step=0)

    if optimizer is not None:
      _ = optimizer.iterations
      optimizer._create_hypers()  # pylint: disable=protected-access
      optimizer._create_slots(self.trainable_variables)  # pylint: disable=protected-access

  def transfer_weights(self, new_model, new_optimizer=None, optimizer=None, ignore_weights=None):
    """Transfers weights (and optionally optimizer slots) from this model to
    another.

    This default implementation assumes that :obj:`self` and :obj:`new_model`
    have exactly the same variables. Subclasses can override this method to
    transfer weights to another model type or architecture. For example,
    :class:`opennmt.models.SequenceToSequence` can transfer weights to a model
    with a different vocabulary.

    All model and optimizer variables are expected to be initialized.

    Args:
      new_model: The new model to transfer weights to.
      new_optimizer: The new optimizer.
      optimizer: The optimizer used for the current model.
      ignore_weights: Optional list of weights to not transfer.
    """
    if type(self) is not type(new_model):
      raise ValueError("Transferring weights to another model type is not supported")
    if ignore_weights is None:
      ignore_weights = set()
    ignore_weights_ref = set(weight.ref() for weight in ignore_weights)
    weights = self.weights
    new_weights = new_model.weights
    for weight, new_weight in zip(weights, new_weights):
      if new_weight.ref() not in ignore_weights_ref:
        new_weight.assign(weight)
        if new_optimizer is not None and optimizer is not None:
          for slot_name in new_optimizer.get_slot_names():
            if slot_name not in optimizer.get_slot_names():
              continue
            new_slot = new_optimizer.get_slot(new_weight, slot_name)
            slot = optimizer.get_slot(weight, slot_name)
            new_slot.assign(slot)

  def map_v1_weights(self, weights):
    """Maps current weights to V1 weights.

    Args:
      weights: A nested dictionary following the scope names used in V1. The
        leaves are tuples with the variable value and optionally the optimizer
        slots.

    Returns:
      A list of tuples associating variables and their V1 equivalent.
    """
    raise NotImplementedError("This model can not restore V1 checkpoints")

  def export_assets(self, asset_dir):
    """Exports additional assets used by this model.

    Args:
      asset_dir: The directory where assets can be written.

    Returns:
      A dictionary of additional assets.
    """
    return self.examples_inputter.export_assets(asset_dir)

  def visualize(self, log_dir):
    """Setups model visualization (e.g. word embedding projections).

    Args:
      log_dir: The log directory.
    """
    self.features_inputter.visualize(self, log_dir)
    if not self.unsupervised:
      self.labels_inputter.visualize(self, log_dir)

  def print_prediction(self, prediction, params=None, stream=None):
    """Prints the model prediction.

    Args:
      prediction: The model prediction.
      params: (optional) Dictionary of formatting parameters.
      stream: (optional) The stream to print to.
    """
    _ = params
    print(prediction, file=stream)

  def print_score(self, score, params=None, stream=None):
    """Prints the score result.

    Args:
      score: The score result (output of :meth:`opennmt.models.Model.score`).
      params: (optional) Dictionary of formatting parameters.
      stream: (optional) The stream to print to.
    """
    _ = params
    print(score, file=stream)


class SequenceGenerator(Model):
  """Base class for models generating sequences."""

  @property
  def decoder_inputter(self):
    """The inputter used on the decoder side."""
    return (
        self.labels_inputter if not self.unsupervised
        else self.features_inputter)

  def score(self, features, labels):
    outputs, _ = self(features, labels=labels)
    cross_entropy = tf.nn.sparse_softmax_cross_entropy_with_logits(
        labels["ids_out"], outputs["logits"])
    weights = tf.sequence_mask(labels["length"], dtype=cross_entropy.dtype)
    masked_cross_entropy = cross_entropy * weights
    scores = tf.reduce_sum(masked_cross_entropy, axis=1)
    results = {
        "cross_entropy": cross_entropy,
        "score": scores,
        "tokens": labels["tokens"],
        "length": self.decoder_inputter.get_length(labels, ignore_special_tokens=True)
    }
    if "attention" in outputs:
      results["attention"] = outputs["attention"]
    return results

  def print_score(self, score, params=None, stream=None):
    if params is None:
      params = {}
    length = score["length"]
    tokens = score["tokens"][:length]
    sentence = self.decoder_inputter.tokenizer.detokenize(tokens)
    token_level_scores = None
    attention = None
    if params.get("with_token_level"):
      token_level_scores = score["cross_entropy"][:length]
    if "attention" in score:
      attention = score["attention"][:length]
    alignment_type = params.get("with_alignments")
    sentence = misc.format_translation_output(
        sentence,
        score=score["score"],
        token_level_scores=token_level_scores,
        attention=attention,
        alignment_type=alignment_type)
    misc.print_as_bytes(sentence, stream=stream)



"""Dataset creation and transformations."""

import numpy as np
import tensorflow as tf

from opennmt.utils import misc


def make_datasets(dataset_cls, filenames):
  """Creates instances of :obj:`dataset_cls`.

  Args:
    dataset_cls: A class inheriting from ``tf.data.Dataset``.
    filenames: A list of filenames or a single filename.

  Returns:
    A list of ``tf.data.Dataset`` instances if multiple :obj:`filenames` are
    passed, otherwise a single ``tf.data.Dataset``.

  Raises:
    ValueError: if :obj:`filenames` is empty.
  """
  if not isinstance(filenames, list):
    filenames = [filenames]
  elif not filenames:
    raise ValueError("At least one data file is required")
  datasets = [
      dataset_cls(filename, compression_type="GZIP" if misc.is_gzip_file(filename) else None)
      for filename in filenames]
  if len(datasets) == 1:
    return datasets[0]
  return datasets

def normalize_weights(datasets, weights=None, sizes=None):
  """Returns normalized dataset weights based on datasets sizes.

  Args:
    datasets: A list of ``tf.data.Dataset`` instances.
    weights: An optional list of dataset weights.
    sizes: The size of each dataset, if known.

  Returns:
    A normalized list of weights that can be used as sampling probabilities.

  Raises:
    ValueError: if the length of :obj:`weights` or :obj:`sizes` does not match
      the length of :obj:`datasets`.
  """
  if not datasets:
    return []
  if len(datasets) == 1:
    return [1.]

  if weights is None:
    weights = [1 / len(datasets)] * len(datasets)
  elif len(weights) != len(datasets):
    raise ValueError("Got %d datasets but %d weights" % (len(datasets), len(weights)))

  if sizes is None:
    sizes = [int(get_dataset_size(dataset)) for dataset in datasets]
  elif len(sizes) != len(datasets):
    raise ValueError("Got %d datasets but %d sizes" % (len(datasets), len(sizes)))

  # Weights should be normalized by the dataset size relative to the total size.
  total_size = sum(sizes)
  weights = [weight * (size / total_size) for weight, size in zip(weights, sizes)]

  # Convert weights to probabilities.
  logits = tf.math.log(tf.constant(weights, dtype=tf.float32))
  probabilities = tf.nn.softmax(logits).numpy().tolist()
  return probabilities

def _get_output_shapes(dataset):
  """Returns the outputs shapes of the dataset.

  Args:
    dataset: A ``tf.data.Dataset``.

  Returns:
    A nested structure of ``tf.TensorShape``
  """
  return tf.nest.map_structure(lambda spec: spec.shape, dataset.element_spec)

def get_dataset_size(dataset, batch_size=5000):
  """Get the dataset size.

  Example:

    >>> dataset = tf.data.Dataset.range(5)
    >>> opennmt.data.get_dataset_size(dataset).numpy()
    5

  Args:
    dataset: A dataset.
    batch_size: The batch size to use to improve the scan performance, or
      ``None`` to scan the dataset as-is.

  Returns:
    The dataset size or ``None`` if the dataset is infinite.
  """
  if tf.data.experimental.cardinality(dataset) == tf.data.experimental.INFINITE_CARDINALITY:
    return None
  if batch_size is not None:
    dataset = dataset.batch(batch_size)

  def _reduce_func(count, element):
    element = tf.nest.flatten(element)[0]
    batch_size = tf.shape(element)[0]
    return count + tf.cast(batch_size, count.dtype)

  return dataset.reduce(tf.constant(0, dtype=tf.int64), _reduce_func)

def filter_irregular_batches(multiple):
  """Transformation that filters out batches based on their size.

  Example:

    >>> dataset = tf.data.Dataset.range(10).batch(3)
    >>> dataset = dataset.apply(opennmt.data.filter_irregular_batches(3))
    >>> len(list(iter(dataset)))
    3

  Args:
    multiple: The divisor of the batch size.

  Returns:
    A ``tf.data.Dataset`` transformation.
  """
  if multiple == 1:
    return lambda dataset: dataset

  def _predicate(*x):
    flat = tf.nest.flatten(x)
    batch_size = tf.shape(flat[0])[0]
    return tf.equal(batch_size % multiple, 0)

  return lambda dataset: dataset.filter(_predicate)

def filter_examples_by_length(maximum_features_length=None,
                              maximum_labels_length=None,
                              features_length_fn=None,
                              labels_length_fn=None):
  """Transformation that filters out examples with zero length or length that is
  greater than the configured maximum.

  Example:

    >>> dataset = dataset.apply(opennmt.data.filter_examples_by_length(...))

  Args:
    maximum_features_length: The maximum length or list of maximum lengths of
      the features sequence(s). ``None`` to not constrain the length.
    maximum_labels_length: The maximum length or list of maximum lengths of
      the labels sequence(s). ``None`` to not constrain the length.
    features_length_fn: A function mapping features to a sequence length.
    labels_length_fn: A function mapping labels to a sequence length.

  Returns:
    A ``tf.data.Dataset`` transformation.
  """
  if features_length_fn is None and labels_length_fn is None:
    return lambda dataset: dataset

  def _length_constraints(length, maximum_length):
    # Work with lists of lengths which correspond to the general multi source case.
    if not isinstance(length, list):
      length = [length]
    if not isinstance(maximum_length, list):
      maximum_length = [maximum_length]
    # Unset maximum lengths are set to None (i.e. no constraint).
    maximum_length += [None] * (len(length) - len(maximum_length))
    constraints = []
    for l, maxlen in zip(length, maximum_length):
      constraints.append(tf.greater(l, 0))
      if maxlen is not None:
        constraints.append(tf.less_equal(l, maxlen))
    return constraints

  def _predicate(features, labels):
    cond = []
    features_length = features_length_fn(features) if features_length_fn is not None else None
    labels_length = labels_length_fn(labels) if labels_length_fn is not None else None
    if features_length is not None:
      cond.extend(_length_constraints(features_length, maximum_features_length))
    if labels_length is not None:
      cond.extend(_length_constraints(labels_length, maximum_labels_length))
    return tf.reduce_all(cond)

  return lambda dataset: dataset.filter(_predicate)

def make_cardinality_multiple_of(divisor):
  """Transformation that ensures that the dataset cardinality is a multiple of
  :obj:`divisor`.

  Example:

    >>> dataset = tf.data.Dataset.range(7)
    >>> dataset = dataset.apply(opennmt.data.make_cardinality_multiple_of(10))
    >>> len(list(iter(dataset)))
    10

  Args:
    divisor: The value that should divide the dataset size.

  Returns:
    A ``tf.data.Dataset`` transformation.

  Tip:
    This transformation is useful when training multiple replicas on a finite
    dataset. It ensures that each replica receives a non empty batch in the last
    training iteration.
  """
  if divisor == 1:
    return lambda dataset: dataset

  def _continue_iter(num_consumed, element):
    # Continue iterating if the current element is from the original dataset or
    # if the number of consumed batches is not a multiple of divisor.
    is_original = element[0]
    return tf.math.logical_or(is_original, tf.math.not_equal(num_consumed % divisor, 0))

  def _retrieve_element(num_consumed, element):
    _ = num_consumed
    return element[1]

  def _transform(dataset):
    # Nothing to do for infinite datasets.
    if tf.data.experimental.cardinality(dataset) == tf.data.experimental.INFINITE_CARDINALITY:
      return dataset

    # Concatenate extra batches with a flag.
    extra_batches = dataset.repeat()
    dataset = dataset.map(lambda *x: (tf.constant(True), x))
    extra_batches = extra_batches.map(lambda *x: (tf.constant(False), x))
    dataset = dataset.concatenate(extra_batches)

    # Take all original batches and the number of extra batches required.
    dataset = dataset.enumerate()
    dataset = dataset.apply(tf.data.experimental.take_while(_continue_iter))
    return dataset.map(_retrieve_element)  # Retrieve the element only.

  return _transform

def random_shard(shard_size, dataset_size):
  """Transformation that shards the dataset in a random order.

  Example:

    >>> dataset = tf.data.Dataset.range(6)
    >>> dataset = dataset.apply(opennmt.data.random_shard(2, 6)
    >>> list(dataset.as_numpy_iterator())
    [0, 1, 4, 5, 2, 3]

  Args:
    shard_size: The number of examples in each shard.
    dataset_size: The total number of examples in the dataset.

  Returns:
    A ``tf.data.Dataset`` transformation.
  """
  num_shards = -(-dataset_size // shard_size)  # Ceil division.
  offsets = np.linspace(0, dataset_size, num=num_shards, endpoint=False, dtype=np.int64)

  def _random_shard(dataset):
    sharded_dataset = tf.data.Dataset.from_tensor_slices(offsets)
    sharded_dataset = sharded_dataset.shuffle(num_shards)
    sharded_dataset = sharded_dataset.flat_map(
        lambda offset: dataset.skip(offset).take(shard_size))
    return sharded_dataset

  return _random_shard

def shuffle_dataset(buffer_size, shuffle_shards=True, dataset_size=None):
  """Transformation that shuffles the dataset based on its size.

  Example:

    >>> dataset = tf.data.Dataset.range(6)
    >>> dataset = dataset.apply(opennmt.data.shuffle_dataset(3))
    >>> list(dataset.as_numpy_iterator())
    [2, 3, 1, 0, 4, 5]

  Args:
    buffer_size: The number of elements from which to sample.
    shuffle_shards: When :obj:`buffer_size` is smaller than the dataset size,
      the dataset is first sharded in a random order to add another level of
      shuffling.
    dataset_size: If the dataset size is already known, it can be passed here to
      avoid a slower generic computation of the dataset size later.

  Returns:
    A ``tf.data.Dataset`` transformation.
  """

  def _shuffle(dataset):
    sample_size = buffer_size
    if sample_size < 0 or shuffle_shards:
      total_size = dataset_size
      if total_size is None:
        total_size = get_dataset_size(dataset)
      tf.get_logger().info("Training on %d examples", total_size)
      if sample_size < 0:
        sample_size = total_size
      elif sample_size < total_size:
        dataset = dataset.apply(random_shard(sample_size, total_size))
    dataset = dataset.shuffle(sample_size)
    return dataset

  return _shuffle

def batch_dataset(batch_size, padded_shapes=None):
  """Transformation that batches a dataset.

  Example:

    >>> dataset = dataset.apply(opennmt.data.batch_dataset(...))

  Args:
    batch_size: The batch size.
    padded_shapes: The padded shapes for this dataset. If ``None``, the shapes
      are automatically inferred from the dataset output shapes.

  Returns:
    A ``tf.data.Dataset`` transformation.

  See Also:
    :func:`opennmt.data.batch_sequence_dataset`
  """
  return lambda dataset: dataset.padded_batch(
      batch_size, padded_shapes=padded_shapes or _get_output_shapes(dataset))

def batch_sequence_dataset(batch_size,
                           batch_type="examples",
                           batch_multiplier=1,
                           batch_size_multiple=1,
                           length_bucket_width=None,
                           length_fn=None,
                           padded_shapes=None):
  """Transformation that batches a dataset of sequences.

  This implements an example-based and a token-based batching strategy
  with optional bucketing of sequences.

  Bucketing makes the batches contain sequences of similar lengths to optimize
  the training efficiency. For example, if :obj:`length_bucket_width` is 5,
  sequences will be organized by the following length buckets:

  1 - 5 | 6 - 10 | 11 - 15 | ...

  Then when building the next batch, sequences will be selected from the same
  length bucket.

  If the dataset has parallel elements (e.g. a parallel source and target
  dataset), the element is assigned to the bucket corresponding to the maximum
  length of all parallel elements.

  Example:

    >>> dataset = dataset.apply(opennmt.data.batch_sequence_dataset(...))

  Args:
    batch_size: The batch size.
    batch_type: The training batching strategy to use: can be "examples" or
      "tokens".
    batch_multiplier: The batch size multiplier.
    batch_size_multiple: When :obj:`batch_type` is "tokens", ensure that the
      resulting batch size is a multiple of this value.
    length_bucket_width: The width of the length buckets to select batch
      candidates from. ``None`` to not constrain batch formation.
    length_fn: A function or list of functions (in case of a parallel dataset)
      that take features as argument and return the associated sequence length.
    padded_shapes: The padded shapes for this dataset. If ``None``, the shapes
      are automatically inferred from the dataset output shapes.

  Returns:
    A ``tf.data.Dataset`` transformation.

  Raises:
    ValueError: if :obj:`batch_type` is not one of "examples" or "tokens".
    ValueError: if the number of length functions in :obj:`length_fn` does not
      match the number of parallel elements.

  See Also:
    :func:`opennmt.data.batch_dataset`
  """
  batch_size = batch_size * batch_multiplier

  def _get_bucket_id(features, length_fn):
    default_id = tf.constant(0, dtype=tf.int32)
    if length_fn is None:
      return default_id
    lengths = length_fn(features)
    if lengths is None:
      return default_id
    if not isinstance(lengths, list):
      lengths = [lengths]  # Fallback to the general case of parallel inputs.
    lengths = [length // length_bucket_width for length in lengths]
    return tf.reduce_max(lengths)

  def _key_func(*args):
    length_fns = length_fn
    if length_fns is None:
      length_fns = [None for _ in args]
    elif not isinstance(length_fns, (list, tuple)):
      length_fns = [length_fns]
    if len(length_fns) != len(args):
      raise ValueError("%d length functions were passed but this dataset contains "
                       "%d parallel elements" % (len(length_fns), len(args)))
    # Take the highest bucket id.
    bucket_id = tf.reduce_max([
        _get_bucket_id(features, length_fn)
        for features, length_fn in zip(args, length_fns)])
    return tf.cast(bucket_id, tf.int64)

  def _reduce_func(unused_key, dataset):
    return dataset.apply(batch_dataset(batch_size, padded_shapes=padded_shapes))

  def _window_size_func(key):
    if length_bucket_width > 1:
      key += 1  # For length_bucket_width == 1, key 0 is unassigned.
    size = batch_size // (key * length_bucket_width)
    required_multiple = batch_multiplier * batch_size_multiple
    if required_multiple > 1:
      size = size + required_multiple - size % required_multiple
    return tf.cast(tf.maximum(size, required_multiple), tf.int64)

  if length_bucket_width is None:
    return batch_dataset(batch_size, padded_shapes=padded_shapes)

  if batch_type == "examples":
    return tf.data.experimental.group_by_window(
        _key_func, _reduce_func, window_size=batch_size)
  elif batch_type == "tokens":
    return tf.data.experimental.group_by_window(
        _key_func, _reduce_func, window_size_func=_window_size_func)
  else:
    raise ValueError(
        "Invalid batch type: '{}'; should be 'examples' or 'tokens'".format(batch_type))


def training_pipeline(batch_size,
                      batch_type="examples",
                      batch_multiplier=1,
                      batch_size_multiple=1,
                      process_fn=None,
                      transform_fns=None,
                      length_bucket_width=None,
                      features_length_fn=None,
                      labels_length_fn=None,
                      maximum_features_length=None,
                      maximum_labels_length=None,
                      single_pass=False,
                      num_shards=1,
                      shard_index=0,
                      num_threads=None,
                      dataset_size=None,
                      shuffle_buffer_size=None,
                      prefetch_buffer_size=None,
                      cardinality_multiple=1):
  """Transformation that applies most of the dataset operations commonly used
  for training on sequence data:

  * sharding
  * shuffling
  * processing
  * filtering
  * bucketization
  * batching
  * prefetching

  Example:

    >>> dataset = dataset.apply(opennmt.data.training_pipeline(...))

  Args:
    batch_size: The batch size to use.
    batch_type: The training batching stragety to use: can be "examples" or
      "tokens".
    batch_multiplier: The batch size multiplier.
    batch_size_multiple: When :obj:`batch_type` is "tokens", ensure that the
      resulting batch size is a multiple of this value.
    process_fn: The processing function to apply on each element.
    transform_fns: List of dataset transformation functions (applied after
      :obj:`process_fn` if defined).
    length_bucket_width: The width of the length buckets to select batch
      candidates from. ``None`` to not constrain batch formation.
    features_length_fn: A function mapping features to a sequence length.
    labels_length_fn: A function mapping labels to a sequence length.
    maximum_features_length: The maximum length or list of maximum lengths of
      the features sequence(s). ``None`` to not constrain the length.
    maximum_labels_length: The maximum length of the labels sequence.
      ``None`` to not constrain the length.
    single_pass: If ``True``, makes a single pass over the training data.
    num_shards: The number of data shards (usually the number of workers in a
      distributed setting).
    shard_index: The shard index this data pipeline should read from.
    num_threads: The number of elements processed in parallel.
    dataset_size: If the dataset size is already known, it can be passed here to
      avoid a slower generic computation of the dataset size later.
    shuffle_buffer_size: The number of elements from which to sample.
    prefetch_buffer_size: The number of batches to prefetch asynchronously. If
      ``None``, use an automatically tuned value.
    cardinality_multiple: Ensure that the dataset cardinality is a multiple of
      this value when :obj:`single_pass` is ``True``.

  Returns:
    A ``tf.data.Dataset`` transformation.

  See Also:
    - :func:`opennmt.data.batch_sequence_dataset`
    - :func:`opennmt.data.make_cardinality_multiple_of`
    - :func:`opennmt.data.filter_examples_by_length`
    - :func:`opennmt.data.filter_irregular_batches`
    - :func:`opennmt.data.shuffle_dataset`
  """
  if dataset_size is not None and num_shards > 1:
    # Update dataset_size based on the shard size.
    if isinstance(dataset_size, list):
      dataset_size = [size // num_shards for size in dataset_size]
    else:
      dataset_size //= num_shards

  def _make_weighted_dataset(datasets, weights):
    if single_pass:
      raise ValueError("single_pass parameter is not compatible with weighted datasets")
    if not datasets:
      raise ValueError("At least one dataset is required")
    if weights is not None and len(weights) != len(datasets):
      raise ValueError("%d dataset weights were provided, but %d were expected to match the "
                       "number of data files" % (len(weights), len(datasets)))
    if num_shards > 1:
      datasets = [dataset.shard(num_shards, shard_index) for dataset in datasets]
    weights = normalize_weights(datasets, weights=weights, sizes=dataset_size)
    datasets = [dataset.repeat() for dataset in datasets]
    dataset = tf.data.experimental.sample_from_datasets(datasets, weights=weights)
    if shuffle_buffer_size is not None and shuffle_buffer_size != 0:
      if shuffle_buffer_size < 0:
        raise ValueError("shuffle_buffer_size < 0 is not compatible with weighted datasets")
      dataset = dataset.shuffle(shuffle_buffer_size)
    return dataset

  def _make_single_dataset(dataset):
    if num_shards > 1:
      dataset = dataset.shard(num_shards, shard_index)
    if shuffle_buffer_size is not None and shuffle_buffer_size != 0:
      dataset = dataset.apply(shuffle_dataset(shuffle_buffer_size, dataset_size=dataset_size))
    return dataset

  def _pipeline(dataset):
    if isinstance(dataset, tuple):
      dataset, weights = dataset
    else:
      weights = None
    is_weighted_dataset = isinstance(dataset, list)
    if is_weighted_dataset:
      dataset = _make_weighted_dataset(dataset, weights)
    else:
      dataset = _make_single_dataset(dataset)
    if process_fn is not None:
      dataset = dataset.map(process_fn, num_parallel_calls=num_threads or 4)
    if transform_fns is not None:
      for transform_fn in transform_fns:
        dataset = dataset.apply(transform_fn)
    dataset = dataset.apply(filter_examples_by_length(
        maximum_features_length=maximum_features_length,
        maximum_labels_length=maximum_labels_length,
        features_length_fn=features_length_fn,
        labels_length_fn=labels_length_fn))
    dataset = dataset.apply(batch_sequence_dataset(
        batch_size,
        batch_type=batch_type,
        batch_multiplier=batch_multiplier,
        batch_size_multiple=batch_size_multiple,
        length_bucket_width=length_bucket_width,
        length_fn=[features_length_fn, labels_length_fn]))
    dataset = dataset.apply(filter_irregular_batches(batch_multiplier))
    if not single_pass:
      if not is_weighted_dataset:  # Weighted dataset is repeated before sampling.
        dataset = dataset.repeat()
    else:
      dataset = dataset.apply(make_cardinality_multiple_of(cardinality_multiple))
    dataset = dataset.prefetch(prefetch_buffer_size)
    return dataset

  return _pipeline

def inference_pipeline(batch_size,
                       process_fn=None,
                       transform_fns=None,
                       length_bucket_width=None,
                       length_fn=None,
                       num_threads=None,
                       prefetch_buffer_size=None):
  """Transformation that applies dataset operations for inference.

  Example:

    >>> dataset = dataset.apply(opennmt.data.inference_pipeline(...))

  Args:
    batch_size: The batch size to use.
    process_fn: The processing function to apply on each element.
    transform_fns: List of dataset transformation functions (applied after
      :obj:`process_fn` if defined).
    length_bucket_width: The width of the length buckets to select batch
      candidates from. If set, this means the inference pipeline will be
      reordered based on the examples length, the application is then
      responsible to restore the predictions in order. An "index" key will be
      inserted in the examples dictionary.
    length_fn: A function mapping features to a sequence length.
    num_threads: The number of elements processed in parallel.
    prefetch_buffer_size: The number of batches to prefetch asynchronously. If
      ``None``, use an automatically tuned value.

  Returns:
    A ``tf.data.Dataset`` transformation.

  Raises:
    ValueError: if :obj:`length_bucket_width` is set but not :obj:`length_fn`.
    ValueError: if :obj:`length_bucket_width` is set but the dataset does not
      output a dictionary structure.
  """

  def _inject_index(index, x):
    x["index"] = index
    return x

  def _pipeline(dataset):
    if process_fn is not None:
      dataset = dataset.map(process_fn, num_parallel_calls=num_threads)
    if transform_fns is not None:
      for transform_fn in transform_fns:
        dataset = dataset.apply(transform_fn)
    if length_bucket_width is not None and length_bucket_width > 0:
      if length_fn is None:
        raise ValueError("length_fn is required when reordering by length")
      if not isinstance(_get_output_shapes(dataset), dict):
        raise ValueError("Reordering by length expects dataset elements to be Python dicts")
      dataset = dataset.enumerate()
      dataset = dataset.map(_inject_index)
      dataset = dataset.apply(batch_sequence_dataset(
          batch_size,
          length_bucket_width=length_bucket_width,
          length_fn=length_fn))
    else:
      dataset = dataset.apply(batch_dataset(batch_size))
    dataset = dataset.prefetch(prefetch_buffer_size)
    return dataset

  return _pipeline

def function_on_next(dataset, as_numpy=False):
  """Decorator to run a ``tf.function`` on each dataset element.

  The motivation for this construct is to get the next element within the
  ``tf.function`` for increased efficiency.

  Args:
    dataset: The dataset to iterate.
    as_numpy: If ``True``, call `.numpy()` on each output tensors.

  Returns:
    A function decorator. The decorated function is transformed into a callable
    that returns a generator over its outputs.
  """

  def decorator(func):
    def _fun():
      iterator = iter(dataset)

      @tf.function
      def _tf_fun():
        return func(lambda: next(iterator))

      while True:
        try:
          outputs = _tf_fun()
          if as_numpy:
            outputs = tf.nest.map_structure(lambda x: x.numpy(), outputs)
          yield outputs
        except tf.errors.OutOfRangeError:
          break

    return _fun

  return decorator



"""Training related classes and functions."""

import abc
import contextlib
import time

import tensorflow as tf

from opennmt.optimizers import utils as optimizer_util
from opennmt.utils import misc


class Trainer(abc.ABC):
  """Base class for model trainer."""

  def __init__(self, model, optimizer, checkpoint=None, is_master=True):
    """Initializes the trainer.

    Args:
      checkpoint: A :class:`opennmt.utils.Checkpoint` instance.
      is_master: Whether this trainer instance is the master trainer.
    """
    self._checkpoint = checkpoint
    self._is_master = is_master
    self._model = model
    if checkpoint is not None:
      self._summary_writer = tf.summary.create_file_writer(checkpoint.model_dir)
    else:
      self._summary_writer = tf.summary.create_noop_writer()
    self._words_counters = {}
    self._gradient_accumulator = optimizer_util.GradientAccumulator()

    if optimizer is None:
      raise ValueError("No optimizer is defined")
    graph_optimizer_options = tf.config.optimizer.get_experimental_options()
    mixed_precision_enabled = graph_optimizer_options.get("auto_mixed_precision")
    if (mixed_precision_enabled
        and not isinstance(optimizer, tf.keras.mixed_precision.experimental.LossScaleOptimizer)):
      optimizer = _LossScaleOptimizer(optimizer, "dynamic")
    self._optimizer = optimizer

  @property
  def num_replicas(self):
    """Number of synchronous training replicas."""
    return 1

  def __call__(self,
               dataset,
               max_step=None,
               accum_steps=1,
               report_steps=100,
               save_steps=5000,
               evaluator=None,
               eval_steps=5000,
               moving_average_decay=None):
    """Runs the training.

    Args:
      dataset: A ``tf.data.Dataset`` or a function taking a ``tf.distribute.InputContext``
        instance and returning a ``tf.data.Dataset``.
      max_step: The final training step.
      accum_steps: The number of gradient accumulation steps.
      report_steps: Report status every this many steps.
      save_steps: Save a checkpoint every this many steps.
      evaluator: A :class:`opennmt.evaluation.Evaluator` instance to call for
        evaluation.
      eval_steps: Evaluate every this many steps.
      moving_average_decay: If set, maintain an exponential moving average of the model
        variables using this decay value (usually close to 1, e.g. 0.9999). See
        https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage.
    """
    if max_step is not None and self._optimizer.iterations.numpy() >= max_step:
      tf.get_logger().warning("Model already reached max_step = %d. Exiting.", max_step)
      return
    if evaluator is not None and evaluator.should_stop():
      tf.get_logger().warning("Early stopping conditions are already met. Exiting.")
      return

    self._gradient_accumulator.reset()
    self._words_counters.clear()

    with self._summary_writer.as_default():
      iterations = self._optimizer.iterations
      tf.summary.experimental.set_step(iterations)

      step = None
      moving_average = None
      last_report_step = iterations.numpy()
      last_report_time = time.time()
      for loss in self._steps(dataset, accum_steps=accum_steps, report_steps=report_steps):
        if tf.math.is_nan(loss):
          raise RuntimeError("Model diverged with loss = NaN.")

        if moving_average_decay is not None and self._is_master:
          if moving_average is None:
            moving_average = MovingAverage(
                self._model.trainable_variables,
                iterations,
                decay=moving_average_decay)
          else:
            moving_average.update()

        step = iterations.numpy()
        if step % report_steps == 0:
          words_counters = self._get_words_counters()
          if self._is_master:
            _report_training_status(
                step,
                loss,
                self._optimizer.learning_rate,
                words_counters,
                last_report_step,
                last_report_time)
            last_report_step = step
            last_report_time = time.time()
        if step == 1 or (save_steps is not None and step % save_steps == 0):
          self._save_checkpoint(step, moving_average=moving_average)
        if eval_steps is not None and step % eval_steps == 0:
          early_stop = self._evaluate(evaluator, step, moving_average=moving_average)
          if early_stop:
            tf.get_logger().warning("Early stopping conditions are met. Exiting.")
            break
        if step == max_step:
          break

      if step is None:
        raise RuntimeError("No training steps were executed. This usually means the "
                           "training file is empty or all examples were filtered out. "
                           "For the latter, verify that the maximum_*_length values are "
                           "consistent with your data.")
      self._save_checkpoint(step, moving_average=moving_average)
      self._evaluate(evaluator, step, moving_average=moving_average)

  def _save_checkpoint(self, step, moving_average=None):
    """Saves a checkpoint for step."""
    if not self._is_master or self._checkpoint is None or step == self._checkpoint.last_saved_step:
      return
    with moving_average.shadow_variables() if moving_average is not None else contextlib.suppress():
      self._checkpoint.save(step)

  def _evaluate(self, evaluator, step, moving_average=None):
    """Runs evaluation for step. Returns ``True`` is early conditions are met."""
    if not self._is_master or evaluator is None or step == evaluator.last_evaluated_step:
      return False
    with moving_average.shadow_variables() if moving_average is not None else contextlib.suppress():
      evaluator(step)
      return evaluator.should_stop()

  @abc.abstractmethod
  def _steps(self, dataset, accum_steps=1, report_steps=None):
    """Returns a generator over training steps (i.e. parameters update).

    Args:
      dataset: The training dataset.
      accum_steps: Accumulate the gradients of this many steps/batches.
      report_steps: Report summary statistics every this many steps. This should
        typically be used in a ``tf.summary.record_if`` context.

    Returns:
      A generator that yields a loss value to report for this step.
    """
    raise NotImplementedError()

  def _run_model(self, source, target):
    """Computes the loss of the given source and target pair.

    Args:
      source: A nested structure of tensors.
      target: A nested structure of tensors.

    Returns:
      A tuple containing,

      - The loss to compute the gradients.
      - The loss to report.
    """
    first_call = not self._model.built
    outputs, _ = self._model(
        source,
        labels=target,
        training=True,
        step=self._optimizer.iterations)
    loss = self._model.compute_loss(outputs, target, training=True)
    if isinstance(loss, tuple):
      training_loss = loss[0] / loss[1]
      reported_loss = loss[0] / loss[2] if len(loss) > 2 else training_loss
    else:
      training_loss, reported_loss = loss, loss
    training_loss = self._model.regularize_loss(
        training_loss, variables=self._model.trainable_variables)
    self._update_words_counter("source", source)
    if not self._model.unsupervised:
      self._update_words_counter("target", target)
    if first_call and self._is_master:
      if self._checkpoint is not None:
        self._model.visualize(self._checkpoint.model_dir)
      tf.get_logger().info("Number of model parameters: %d", self._model.count_params())
      tf.get_logger().info(
          "Number of model weights: %d (trainable = %d, non trainable = %d)",
          len(self._model.weights),
          len(self._model.trainable_weights),
          len(self._model.non_trainable_weights))
    return training_loss, reported_loss

  def _should_record_summaries(self, report_steps):
    """Returns a boolean tensor to be used in tf.summary.record_if."""
    if report_steps is None or not self._is_master:
      return False
    else:
      return tf.logical_and(
          tf.equal(self._optimizer.iterations % report_steps, 0),
          tf.equal(self._gradient_accumulator.step, 0))

  def _forward(self, source, target, record_summaries=False):
    """Forwards a training example and accumulates the gradients."""
    with tf.summary.record_if(record_summaries):
      training_loss, reported_loss = self._run_model(source, target)
      gradients = self._optimizer.get_gradients(training_loss, self._model.trainable_variables)
      self._gradient_accumulator(gradients)
      _summarize_gradients(gradients, record_summaries)
    return reported_loss

  def _step(self):
    """Applies gradients and resets accumulation."""
    gradient_scale = self._gradient_accumulator.step * self.num_replicas
    gradients = [
        self._all_reduce_sum(gradient / tf.cast(gradient_scale, gradient.dtype))
        for gradient in self._gradient_accumulator.gradients]
    self._optimizer.apply_gradients(list(zip(gradients, self._model.trainable_variables)))
    self._gradient_accumulator.reset()

  def _update_words_counter(self, name, features):
    """Accumulates number of source and target tokens to report throughput."""
    length = features.get("length")
    if length is None:
      return
    num_words = tf.reduce_sum(length)
    counter = self._words_counters.get(name)
    if counter is None:
      counter = tf.Variable(
          tf.constant(0, dtype=tf.int64),
          trainable=False,
          synchronization=tf.VariableSynchronization.ON_READ,
          aggregation=tf.VariableAggregation.SUM)
      self._words_counters[name] = counter
    counter.assign_add(tf.cast(num_words, tf.int64))

  @tf.function
  def _get_words_counters(self):
    """Returns the accumulated words counters and resets them.

    This is used to report the words per second in the training logs.

    Returns:
      A dictionary mapping a counter name to a Python value.
    """
    counters = {}
    for name, counter in self._words_counters.items():
      counters[name] = self._all_reduce_sum(counter.read_value())
      counter.assign(tf.constant(0, dtype=tf.int64))
    return counters

  def _all_reduce_sum(self, value):
    """Reduces the value across all replicas."""
    return value


class BasicTrainer(Trainer):
  """Basic single GPU trainer."""

  def _steps(self, dataset, accum_steps=1, report_steps=None):
    if accum_steps != 1:
      raise ValueError("BasicTrainer does not support gradient accumulation")
    if callable(dataset):
      dataset = dataset(tf.distribute.InputContext())

    @tf.function(input_signature=dataset.element_spec)
    def _step(source, target):
      training_loss, reported_loss = self._run_model(source, target)
      variables = self._model.trainable_variables
      gradients = self._optimizer.get_gradients(training_loss, variables)
      self._optimizer.apply_gradients(list(zip(gradients, variables)))
      return reported_loss

    for source, target in dataset:
      yield _step(source, target)


class HorovodTrainer(Trainer):
  """Trainer compatible with Horovod distributed training."""

  def __init__(self, model, optimizer, hvd, checkpoint=None):
    super().__init__(model, optimizer, checkpoint=checkpoint, is_master=hvd.rank() == 0)
    self._hvd = hvd

  @property
  def num_replicas(self):
    return self._hvd.size()

  def _all_reduce_sum(self, value):
    return self._hvd.allreduce(value, op=self._hvd.Sum)

  def _steps(self, dataset, accum_steps=1, report_steps=None):
    if callable(dataset):
      dataset = dataset(tf.distribute.InputContext(
          num_input_pipelines=self._hvd.size(),
          input_pipeline_id=self._hvd.rank(),
          num_replicas_in_sync=self._hvd.size()))

    # Wrap forward and step with tf.function.

    @tf.function(input_signature=dataset.element_spec)
    def _forward(source, target):
      return self._forward(
          source,
          target,
          record_summaries=self._should_record_summaries(report_steps))

    @tf.function
    def _step():
      return self._step()

    for i, (source, target) in enumerate(dataset):
      loss = _forward(source, target)
      _assert_loss_is_finite(loss)
      if i == 0 or (i + 1) % accum_steps == 0:
        _step()
        if i == 0:
          self._hvd.broadcast_variables(self._model.variables, root_rank=0)
          self._hvd.broadcast_variables(self._optimizer.variables(), root_rank=0)
        yield loss


class DistributionStrategyTrainer(Trainer):
  """Trainer based on distribution strategies."""

  def __init__(self, model, optimizer, checkpoint=None, devices=None):
    """Initializes the trainer.

    Args:
      checkpoint: A :class:`opennmt.utils.checkpoint.Checkpoint` instance.
      devices: List of device strings to use for training.
    """
    super(DistributionStrategyTrainer, self).__init__(model, optimizer, checkpoint=checkpoint)
    if not devices:
      devices = misc.get_devices(count=1)  # Train with 1 device by default.
    self._strategy = tf.distribute.MirroredStrategy(devices=devices)
    with self._strategy.scope():
      # Create some variables under the strategy scope.
      _ = self._optimizer.iterations

  @property
  def num_replicas(self):
    return self._strategy.num_replicas_in_sync

  def _steps(self, dataset, accum_steps=1, report_steps=None):
    for i, loss in enumerate(self._accumulate_next_gradients(dataset, report_steps=report_steps)):
      _assert_loss_is_finite(loss)
      if i == 0 or (i + 1) % accum_steps == 0:
        self._apply_gradients()
        yield loss

  def _accumulate_next_gradients(self, dataset, report_steps=None):
    """Accumulates the gradients from the next element in :obj:`dataset`."""

    # We prefer not to use experimental_distribute_dataset here because it
    # sometimes fails to split the batches (noticed with tokens batch type).
    # We also assume for now that we are training with a single worker
    # otherwise we would need to correctly shard the input dataset.
    dataset_fn = dataset if callable(dataset) else lambda _: dataset
    distributed_dataset = self._strategy.experimental_distribute_datasets_from_function(
        dataset_fn)

    # Get the next element within the tf.function for more pipelining.
    # See: https://github.com/tensorflow/tensorflow/issues/29075#issuecomment-513390242
    iterator = iter(distributed_dataset)

    @tf.function
    def _accumulate_next():
      record_summaries = self._should_record_summaries(report_steps)
      per_replica_source, per_replica_target = next(iterator)
      return self._accumulate_gradients(
          per_replica_source,
          per_replica_target,
          record_summaries=record_summaries)

    while True:
      try:
        yield _accumulate_next()
      except tf.errors.OutOfRangeError:
        break

  def _accumulate_gradients(self,
                            per_replica_source,
                            per_replica_target,
                            record_summaries=False):
    """Accumulates the gradients (cross-replica)."""
    per_replica_loss = self._strategy.run(
        self._forward,
        args=(per_replica_source, per_replica_target),
        kwargs=dict(record_summaries=record_summaries))
    # TODO: this reduction could be delayed until _step is called.
    return self._strategy.reduce(tf.distribute.ReduceOp.MEAN, per_replica_loss, None)

  @tf.function
  def _apply_gradients(self):
    """Applies the gradients (cross-replica)."""
    self._strategy.run(self._step)


def _report_training_status(step,
                            loss,
                            learning_rate,
                            words_counters,
                            last_report_step,
                            last_report_time):
  elapsed_time = time.time() - last_report_time

  steps_per_sec = (step - last_report_step) / elapsed_time
  tf.summary.scalar("steps_per_sec", steps_per_sec, description="Training steps per second")
  steps_per_sec_fmt = "steps/s = %0.2f" % steps_per_sec

  words_per_sec_fmt = []
  for name, counter in words_counters.items():
    avg = int(counter.numpy() / elapsed_time)
    tf.summary.scalar(
        "words_per_sec/%s" % name,
        avg,
        description="%s words per second" % name.capitalize())
    words_per_sec_fmt.append("%s words/s = %d" % (name, avg))

  if isinstance(learning_rate, tf.optimizers.schedules.LearningRateSchedule):
    learning_rate = learning_rate(step)
  elif isinstance(learning_rate, tf.Variable):
    learning_rate = learning_rate.value()

  tf.get_logger().info(
      "Step = %d ; %s ; Learning rate = %f ; Loss = %f",
      step,
      ", ".join([steps_per_sec_fmt] + list(sorted(words_per_sec_fmt))),
      learning_rate,
      loss)
  tf.summary.scalar("loss", loss, description="Training loss")
  tf.summary.scalar("optim/learning_rate", learning_rate, description="Learning rate")

def _assert_loss_is_finite(loss):
  if tf.math.is_nan(loss):
    raise RuntimeError("Model diverged with loss = NaN.")

def _summarize_gradients(gradients, should_record):
  # Only compute the gradients global norm when the value is actually recorded.
  if isinstance(should_record, bool) and not should_record:
    return
  tf.summary.scalar(
      "gradients/global_norm",
      tf.cond(
          should_record,
          true_fn=lambda: tf.linalg.global_norm(gradients),
          false_fn=lambda: tf.constant(0, dtype=gradients[0].dtype)))


class MovingAverage(object):
  """Object holding an exponential moving average of variables."""

  def __init__(self, variables, step, decay=0.9999):
    """Initializes the moving average object.

    Args:
      variables: The list of variable for which to maintain a moving average.
      step: The training step counter as a ``tf.Variable``.
      decay: The decay rate of the exponential moving average. Usually close to
        1, e.g. 0.9999, see the complete formula on
        https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage.

    Raises:
      TypeError: is :obj:`step` is not a ``tf.Variable``.
    """
    if not isinstance(step, tf.Variable):
      raise TypeError("step should be a tf.Variable")
    if decay < 0.9 or decay > 1:
      tf.get_logger().warning("Moving average decay should be close to 1 (e.g. 0.9999) but you "
                              "passed %f, is it correct? See https://www.tensorflow.org/api_docs"
                              "/python/tf/train/ExponentialMovingAverage for details about the "
                              "formula and recommended decay values.")
    self._ema = tf.train.ExponentialMovingAverage(decay, num_updates=step)
    self._variables = variables
    self.update()

  @tf.function
  def update(self):
    """Updates the moving average of the variables."""
    self._ema.apply(var_list=list(map(misc.get_primary_variable, self._variables)))

  @contextlib.contextmanager
  def shadow_variables(self):
    """Returns a context manager that assigns the variables to their moving
    average value on enter and restores the previous value on exit.

    Returns:
      A context manager.
    """
    # TODO: Do we want to shadow the values on all replicas?
    previous_values = []
    for variable in self._variables:
      previous_values.append(variable.value())
      variable.assign(self._ema.average(misc.get_primary_variable(variable)))
    yield
    for previous_value, variable in zip(previous_values, self._variables):
      variable.assign(previous_value)


class _LossScaleOptimizer(tf.keras.mixed_precision.experimental.LossScaleOptimizer):
  # TODO: Remove this wrapper when this fix is released:
  # https://github.com/tensorflow/tensorflow/commit/d1dd08dd2807ac80a4508686618419826463374b

  def get_unscaled_gradients(self, grads):
    loss_scale_reciprocal = 1. / self._loss_scale()
    return [
        _multiply_gradient(g, loss_scale_reciprocal) if g is not None else None
        for g in grads
    ]


def _multiply_gradient(gradient, scale):
  """Multiply a (possibly sparse) gradient by the given scale factor."""
  scale = tf.cast(scale, gradient.dtype)
  if isinstance(gradient, tf.IndexedSlices):
    return tf.IndexedSlices(
        gradient.values * scale,
        gradient.indices,
        dense_shape=gradient.dense_shape)
  else:
    return gradient * scale



"""Various utility functions to use throughout the project."""

import collections
import copy
import sys
import functools
import heapq
import os

import numpy as np
import tensorflow as tf

from tensorflow.python.training.tracking import graph_view


def get_devices(count=1, fallback_to_cpu=True):
  """Gets devices.

  Args:
    count: The number of devices to get.
    fallback_to_cpu: If ``True``, return CPU devices if no GPU is available.

  Returns:
    A list of device names.

  Raises:
    ValueError: if :obj:`count` is greater than the number of visible devices.
  """
  devices = tf.config.list_logical_devices(device_type="GPU")
  if not devices and fallback_to_cpu:
    devices = tf.config.list_logical_devices(device_type="CPU")
  if len(devices) < count:
    raise ValueError("Requested %d devices but only %d %s visible" % (
        count, len(devices), "is" if len(devices) == 1 else "are"))
  return devices[0:count]

def get_variables_name_mapping(root, root_key=None):
  """Returns mapping between variables and their name in the object-based
  representation.

  Args:
    root: The root layer.
    root_key: Key that was used to save :obj:`root`, if any.

  Returns:
    A dict mapping variables ref to names and a dict mapping variables name to
    variables.
  """
  # TODO: find a way to implement this function using public APIs.
  named_variables, _, _ = graph_view.ObjectGraphView(root).serialize_object_graph()
  variables_to_names = {}
  names_to_variables = {}
  for saveable_object in named_variables:
    variable = saveable_object.op
    if not hasattr(variable, "ref"):  # Ignore non Tensor-like objects.
      continue
    name = saveable_object.name
    if root_key is not None:
      name = "%s/%s" % (root_key, name)
    variables_to_names[variable.ref()] = name
    names_to_variables[name] = variable
  return variables_to_names, names_to_variables

def get_variable_name(variable, root, model_key="model"):
  """Gets the variable name in the object-based representation."""
  variables_to_names, _ = get_variables_name_mapping(root, root_key=model_key)
  # In case of a MirroredVariable, look up the primary variable
  variable = get_primary_variable(variable)
  return variables_to_names.get(variable.ref())

def get_primary_variable(variable):
  """If :obj:`variable` is distributed, returns the primary component."""
  # TODO: use a public API to get the primary variable.
  for attribute_name in ("primary", "_primary"):
    attribute = getattr(variable, attribute_name, None)
    if attribute is not None:
      return attribute
  return variable

def print_as_bytes(text, stream=None):
  """Prints a string as bytes to non rely on :obj:`stream` default encoding.

  Args:
    text: The text to print.
    stream: The stream to print to (``sys.stdout`` if not set).
  """
  if stream is None:
    stream = sys.stdout
  write_buffer = stream.buffer if hasattr(stream, "buffer") else stream
  write_buffer.write(tf.compat.as_bytes(text))
  write_buffer.write(b"\n")
  stream.flush()

def format_translation_output(sentence,
                              score=None,
                              token_level_scores=None,
                              attention=None,
                              alignment_type=None):
  """Formats a translation output with possibly scores, alignments, etc., e.g:

  1.123214 ||| Hello world ||| 0.30907777 0.030488174 ||| 0-0 1-1

  Args:
    sentence: The translation to output.
    score: If set, attach the score.
    token_level_scores: If set, attach the token level scores.
    attention: The attention vector.
    alignment_type: The type of alignments to format (can be: "hard", "soft").
  """
  if score is not None:
    sentence = "%f ||| %s" % (score, sentence)
  if token_level_scores is not None:
    scores_str = " ".join("%f" % s for s in token_level_scores)
    sentence = "%s ||| %s" % (sentence, scores_str)
  if attention is not None and alignment_type is not None:
    if alignment_type == "hard":
      source_indices = np.argmax(attention, axis=-1)
      target_indices = range(attention.shape[0])
      pairs = ("%d-%d" % (src, tgt) for src, tgt in zip(source_indices, target_indices))
      sentence = "%s ||| %s" % (sentence, " ".join(pairs))
    elif alignment_type == "soft":
      vectors = []
      for vector in attention:
        vectors.append(" ".join("%.6f" % value for value in vector))
      sentence = "%s ||| %s" % (sentence, " ; ".join(vectors))
    else:
      raise ValueError("Invalid alignment type %s" % alignment_type)
  return sentence

def item_or_tuple(x):
  """Returns :obj:`x` as a tuple or its single element."""
  x = tuple(x)
  if len(x) == 1:
    return x[0]
  else:
    return x

def count_lines(filename, buffer_size=65536):
  """Returns the number of lines of the file :obj:`filename`."""
  with tf.io.gfile.GFile(filename, mode="rb") as f:
    num_lines = 0
    while True:
      data = f.read(buffer_size)
      if not data:
        return num_lines
      num_lines += data.count(b"\n")

def is_gzip_file(filename):
  """Returns ``True`` if :obj:`filename` is a GZIP file."""
  return filename.endswith(".gz")

def shape_list(x):
  """Return list of dims, statically where possible."""
  x = tf.convert_to_tensor(x)

  # If unknown rank, return dynamic shape
  if x.shape.dims is None:
    return tf.shape(x)

  static = x.shape.as_list()
  shape = tf.shape(x)

  ret = []
  for i, _ in enumerate(static):
    dim = static[i]
    if dim is None:
      dim = shape[i]
    ret.append(dim)
  return ret

def index_structure(structure, path, path_separator="/"):
  """Follows :obj:`path` in a nested structure of objects, lists, and dicts."""
  keys = path.split(path_separator)
  for i, key in enumerate(keys):
    current_path = "%s%s" % (path_separator, path_separator.join(keys[:i]))
    if isinstance(structure, list):
      try:
        index = int(key)
      except ValueError:
        raise ValueError("Object referenced by path '%s' is a list, but got non "
                         "integer index '%s'" % (current_path, key))
      if index < 0 or index >= len(structure):
        raise ValueError("List referenced by path '%s' has length %d, but got "
                         "out of range index %d" % (current_path, len(structure), index))
      structure = structure[index]
    elif isinstance(structure, dict):
      structure = structure.get(key)
      if structure is None:
        raise ValueError("Dictionary referenced by path '%s' does not have the "
                         "key '%s'" % (current_path, key))
    else:
      structure = getattr(structure, key, None)
      if structure is None:
        raise ValueError("Object referenced by path '%s' does not have the "
                         "attribute '%s'" % (current_path, key))
  return structure

def clone_layer(layer):
  """Clones a layer."""
  return copy.deepcopy(layer)

def set_dropout(root_layer, dropout):
  """Overrides all dropout values in :obj:`root_layer` and its descendants.

  Args:
    dropout: The dropout value to set.

  Raises:
    ValueError: if :obj:`root_layer` is not a ``tf.Module``.
  """
  if not isinstance(root_layer, tf.Module):
    raise ValueError("Layer should be a tf.Module")
  for layer in (root_layer,) + root_layer.submodules:
    for attr, value in layer.__dict__.copy().items():
      if isinstance(value, tf.keras.layers.Dropout):
        value.rate = dropout
      elif "dropout" in attr:
        setattr(layer, attr, dropout)

def extract_batches(tensors):
  """Returns a generator to iterate on each batch of a Numpy array or dict of
  Numpy arrays."""
  if not isinstance(tensors, dict):
    for tensor in tensors:
      yield tensor
  else:
    batch_size = None
    for value in tensors.values():
      batch_size = batch_size or value.shape[0]
    for b in range(batch_size):
      yield {
          key: value[b] for key, value in tensors.items()
      }

def extract_prefixed_keys(dictionary, prefix):
  """Returns a dictionary with all keys from :obj:`dictionary` that are prefixed
  with :obj:`prefix`.
  """
  sub_dict = {}
  for key, value in dictionary.items():
    if key.startswith(prefix):
      original_key = key[len(prefix):]
      sub_dict[original_key] = value
  return sub_dict

def extract_suffixed_keys(dictionary, suffix):
  """Returns a dictionary with all keys from :obj:`dictionary` that are suffixed
  with :obj:`suffix`.
  """
  sub_dict = {}
  for key, value in dictionary.items():
    if key.endswith(suffix):
      original_key = key[:-len(suffix)]
      sub_dict[original_key] = value
  return sub_dict

def merge_dict(dict1, dict2):
  """Merges :obj:`dict2` into :obj:`dict1`.

  Args:
    dict1: The base dictionary.
    dict2: The dictionary to merge.

  Returns:
    The merged dictionary :obj:`dict1`.
  """
  for key, value in dict2.items():
    if isinstance(value, dict):
      dict1[key] = merge_dict(dict1.get(key, {}), value)
    else:
      dict1[key] = value
  return dict1

def read_summaries(event_dir, event_file_pattern="events.out.tfevents.*"):
  """Reads summaries from TensorFlow event files.

  Args:
    event_dir: Directory containing event files.
    event_file_pattern: The pattern to look for event files.

  Returns:
    A list of tuple (step, dict of summaries), sorted by step.
  """
  if not tf.io.gfile.exists(event_dir):
    return []
  summaries = collections.defaultdict(dict)
  for event_file in tf.io.gfile.glob(os.path.join(event_dir, event_file_pattern)):
    for event in tf.compat.v1.train.summary_iterator(event_file):
      if not event.HasField("summary"):
        continue
      for value in event.summary.value:
        tensor_proto = value.tensor
        tensor = tf.io.parse_tensor(
            tensor_proto.SerializeToString(), tf.as_dtype(tensor_proto.dtype))
        summaries[event.step][value.tag] = tf.get_static_value(tensor)
  return list(sorted(summaries.items(), key=lambda x: x[0]))

class OrderRestorer(object):
  """Helper class to restore out-of-order elements in order."""

  def __init__(self, index_fn, callback_fn):
    """Initializes this object.

    Args:
      index_fn: A callable mapping an element to a unique index.
      callback_fn: A callable taking an element that will be called in order.
    """
    self._index_fn = index_fn
    self._callback_fn = callback_fn
    self._next_index = 0
    self._elements = {}
    self._heap = []

  @property
  def buffer_size(self):
    """Number of elements waiting to be notified."""
    return len(self._heap)

  @property
  def next_index(self):
    """The next index to be notified."""
    return self._next_index

  def _try_notify(self):
    old_index = self._next_index
    while self._heap and self._heap[0] == self._next_index:
      index = heapq.heappop(self._heap)
      value = self._elements.pop(index)
      self._callback_fn(value)
      self._next_index += 1
    return self._next_index != old_index

  def push(self, x):
    """Push event :obj:`x`."""
    index = self._index_fn(x)
    if index is None:
      self._callback_fn(x)
      return True
    if index < self._next_index:
      raise ValueError("Event index %d was already notified" % index)
    self._elements[index] = x
    heapq.heappush(self._heap, index)
    return self._try_notify()

class ClassRegistry(object):
  """Helper class to create a registry of classes."""

  def __init__(self, base_class=None):
    """Initializes the class registry.

    Args:
      base_class: Ensure that classes added to this registry are a subclass of
        :obj:`base_class`.
    """
    self._base_class = base_class
    self._registry = {}

  @property
  def class_names(self):
    """Class names registered in this registry."""
    return set(self._registry.keys())

  def register(self, cls=None, name=None, alias=None):
    """Registers a class.

    Args:
      cls: The class to register. If not set, this method returns a decorator for
        registration.
      name: The class name. Defaults to ``cls.__name__``.
      alias: An optional alias or list of alias for this class.

    Returns:
      :obj:`cls` if set, else a class decorator.

    Raises:
      TypeError: if :obj:`cls` does not extend the expected base class.
      ValueError: if the class name is already registered.
    """
    if cls is None:
      return functools.partial(self.register, name=name, alias=alias)
    if self._base_class is not None and not issubclass(cls, self._base_class):
      raise TypeError("Class %s does not extend %s" % (cls.__name__, self._base_class.__name__))
    if name is None:
      name = cls.__name__
    self._register(cls, name)
    if alias is not None:
      if not isinstance(alias, (list, tuple)):
        alias = (alias,)
      for alias_name in alias:
        self._register(cls, alias_name)
    return cls

  def _register(self, cls, name):
    if name in self._registry:
      raise ValueError("Class name %s is already registered" % name)
    self._registry[name] = cls

  def get(self, name):
    """Returns the class with name :obj:`name` or ``None`` if it does not exist
    in the registry.
    """
    return self._registry.get(name)



"""Define model exporters."""

import abc
import os
import tempfile

import tensorflow as tf

from opennmt.utils import misc


def make_exporter(name, **kwargs):
  """Creates a new exporter.

  Args:
    name: The exporter name, can be "saved_model", "ctranslate2".
    **kwargs: Additional arguments to pass to the exporter constructor.

  Returns:
    A :class:`opennmt.utils.Exporter` instance.

  Raises:
    ValueError: if :obj:`name` is invalid.
  """
  if name == "saved_model":
    return SavedModelExporter(**kwargs)
  elif name == "ctranslate2":
    return CTranslate2Exporter(**kwargs)
  else:
    raise ValueError("Invalid exporter name: %s" % name)


class Exporter(abc.ABC):
  """Base class for model exporters."""

  def export(self, model, export_dir):
    """Exports :obj:`model` to :obj:`export_dir`.

    Raises:
      ValueError: if :obj:`model` is not supported by this exporter.
    """
    self._export_model(model, export_dir)
    with tempfile.TemporaryDirectory() as tmp_dir:
      extra_assets = model.export_assets(tmp_dir)
      if extra_assets:
        assets_extra = os.path.join(export_dir, "assets.extra")
        tf.io.gfile.makedirs(assets_extra)
        for filename, path in extra_assets.items():
          tf.io.gfile.copy(path, os.path.join(assets_extra, filename), overwrite=True)
        tf.get_logger().info("Extra assets written to: %s", assets_extra)

  @abc.abstractmethod
  def _export_model(self, model, export_dir):
    raise NotImplementedError()


class SavedModelExporter(Exporter):
  """SavedModel exporter."""

  def _export_model(self, model, export_dir):
    tf.saved_model.save(model, export_dir, signatures=model.serve_function())


class CTranslate2Exporter(Exporter):
  """CTranslate2 exporter."""

  def __init__(self, quantization=None):
    """Initializes the exporter.

    Args:
      quantization: Quantize model weights to this type when exporting the model.
        Can be "int16" or "int8". Default is no quantization.
    """
    # Fail now if ctranslate2 package is missing.
    import ctranslate2  # pylint: disable=import-outside-toplevel,unused-import
    self._quantization = quantization

  def _export_model(self, model, export_dir):
    model_spec = model.ctranslate2_spec
    if model_spec is None:
      raise ValueError("The model does not define an equivalent CTranslate2 model specification")
    if not model.built:
      model.create_variables()
    _, variables = misc.get_variables_name_mapping(model, root_key="model")
    variables = {
        name.replace("/.ATTRIBUTES/VARIABLE_VALUE", ""):value.numpy()
        for name, value in variables.items()}
    import ctranslate2  # pylint: disable=import-outside-toplevel
    converter = ctranslate2.converters.OpenNMTTFConverter(
        src_vocab=model.features_inputter.vocabulary_file,
        tgt_vocab=model.labels_inputter.vocabulary_file,
        variables=variables)
    converter.convert(export_dir, model_spec, quantization=self._quantization, force=True)



"""Hypotheses file scoring."""
import numpy
import pyter

def wer(ref_path, hyp_path):
  """ Compute Word Error Rate between two files """
  with open(ref_path) as ref_fp, open(hyp_path) as hyp_fp:
    ref_line = ref_fp.readline()
    hyp_line = hyp_fp.readline()
    wer_score = 0.0
    line_cpt = 0.0
    while ref_line and hyp_line:
      wer_score = wer_score+(pyter.edit_distance(ref_line.strip().split(), \
                  hyp_line.strip().split())/len(ref_line.strip().split()))
      line_cpt = line_cpt+1
      ref_line = ref_fp.readline()
      hyp_line = hyp_fp.readline()
  mean_wer = 1.0
  if line_cpt > 0:
    mean_wer = wer_score/line_cpt
  return mean_wer

def sentence_wer(ref_sent, hyp_sent):
  """ Compute Word Error Rate between two sentences (as list of words) """
  mwer = numpy.zeros((len(ref_sent)+1)*(len(hyp_sent)+1), \
         dtype=numpy.uint8).reshape((len(ref_sent)+1, len(hyp_sent)+1))
  for i in range(len(ref_sent)+1):
    mwer[i][0] = i
  for j in range(len(hyp_sent)+1):
    mwer[0][j] = j
  for i in range(1, len(ref_sent)+1):
    for j in range(1, len(hyp_sent)+1):
      if ref_sent[i-1] == hyp_sent[j-1]:
        mwer[i][j] = mwer[i-1][j-1]
      else:
        substitute = mwer[i-1][j-1] + 1
        insert = mwer[i][j-1] + 1
        delete = mwer[i-1][j] + 1
        mwer[i][j] = min(substitute, insert, delete)
  sent_wer = 1.0
  if len(ref_sent) > 0:
    sent_wer = mwer[len(ref_sent)][len(hyp_sent)]/len(ref_sent)
  return sent_wer



"""Hypotheses file scoring for Precision Recall and F-Measure."""

def fmeasure(ref_path,
             hyp_path,
             return_precision_only=False,
             return_recall_only=False,
             return_fmeasure_only=False):
  """Compute Precision Recall and F-Measure between two files"""
  with open(ref_path) as ref_fp, open(hyp_path) as hyp_fp:
    list_null_tags = ["X", "null", "NULL", "Null", "O"]
    listtags = []
    classref = []
    classrandom = []
    classhyp = []
    nbrtagref = {}
    nbrtaghyp = {}
    nbrtagok = {}
    for line in ref_fp:
      line = line.strip()
      tabline = line.split(' ')
      lineref = []
      for tag in tabline:
        lineref.append(tag)
        if tag in nbrtagref.keys() and tag not in list_null_tags:
          nbrtagref[tag] = nbrtagref[tag]+1
        else:
          nbrtagref[tag] = 1
      classref.append(lineref)
    for line, lineref in zip(hyp_fp, classref):
      line = line.strip()
      tabline = line.split(' ')
      linehyp = []
      linerandom = []
      for tagcpt, tag in enumerate(tabline):
        linehyp.append(tag)
        if tag not in listtags:
          listtags.append(tag)
        linerandom.append(tag)
        if tagcpt < len(lineref) and tag == lineref[tagcpt]:
          if tag in nbrtagok.keys():
            nbrtagok[tag] = nbrtagok[tag]+1
          else:
            nbrtagok[tag] = 1
        if tag in nbrtaghyp.keys():
          nbrtaghyp[tag] = nbrtaghyp[tag]+1
        else:
          nbrtaghyp[tag] = 1
      classhyp.append(linehyp)
      classrandom.append(linerandom)

  fullprecision = 0
  fullrecall = 0
  precision = {}
  recall = {}
  fulltagok = 0.00
  fulltaghyp = 0.00
  fulltagref = 0.00
  for tag in listtags:
    if tag not in nbrtagok:
      nbrtagok[tag] = 0
    if tag not in nbrtaghyp:
      nbrtaghyp[tag] = 0
    if tag not in nbrtagref:
      nbrtagref[tag] = 0
    if nbrtaghyp[tag] != 0:
      precision[tag] = nbrtagok[tag]/nbrtaghyp[tag]
    else:
      precision[tag] = 0
    if nbrtagref[tag] != 0:
      recall[tag] = nbrtagok[tag]/nbrtagref[tag]
    else:
      recall[tag] = 0
    if tag not in list_null_tags:
      fulltagok = fulltagok+nbrtagok[tag]
      fulltaghyp = fulltaghyp+nbrtaghyp[tag]
      fulltagref = fulltagref+nbrtagref[tag]
  fullprecision = fulltagok / fulltaghyp if fulltaghyp != 0 else 0
  fullrecall = fulltagok / fulltagref if fulltagref != 0 else 0
  fullfmeasure = (
      (2 * fullprecision * fullrecall) / (fullprecision + fullrecall)
      if (fullprecision + fullrecall) != 0 else 0)
  if return_precision_only:
    return fullprecision
  if return_recall_only:
    return fullrecall
  if return_fmeasure_only:
    return fullfmeasure
  return fullprecision, fullrecall, fullfmeasure



"""Hypotheses file scoring."""
import pyter

def ter(ref_path, hyp_path):
  """ Compute Translation Edit Rate between two files """
  with open(ref_path) as ref_fp, open(hyp_path) as hyp_fp:
    ref_line = ref_fp.readline()
    hyp_line = hyp_fp.readline()
    ter_score = 0.0
    line_cpt = 0.0
    while ref_line and hyp_line:
      ter_score = ter_score+(pyter.ter(hyp_line.strip().split(), \
                  ref_line.strip().split()))
      line_cpt = line_cpt+1
      ref_line = ref_fp.readline()
      hyp_line = hyp_fp.readline()
  mean_ter = 1.0
  if line_cpt > 0:
    mean_ter = ter_score/line_cpt
  return mean_ter



"""Hypotheses file scoring."""

import abc

import tensorflow as tf

from rouge import FilesRouge
from sacrebleu import corpus_bleu

from opennmt.utils import misc
from opennmt.utils.fmeasure import fmeasure
from opennmt.utils.ter import ter
from opennmt.utils.wer import wer


class Scorer(abc.ABC):
  """Scores hypotheses against references."""

  def __init__(self, name):
    self._name = name

  @property
  def name(self):
    """The scorer name."""
    return self._name

  @property
  def scores_name(self):
    """The names of returned scores."""
    return {self._name}

  @abc.abstractmethod
  def __call__(self, ref_path, hyp_path):
    """Scores hypotheses.

    Args:
      ref_path: Path to the reference.
      hyp_path: Path to the hypotheses.

    Returns:
      The score or dictionary of scores.
    """
    raise NotImplementedError()

  def lower_is_better(self):
    """Returns ``True`` if a lower score is better."""
    return False

  def higher_is_better(self):
    """Returns ``True`` if a higher score is better."""
    return not self.lower_is_better()


_SCORERS_REGISTRY = misc.ClassRegistry(base_class=Scorer)
register_scorer = _SCORERS_REGISTRY.register  # pylint: disable=invalid-name


@register_scorer(name="rouge")
class ROUGEScorer(Scorer):
  """ROUGE scorer based on https://github.com/pltrdy/rouge."""

  def __init__(self):
    super(ROUGEScorer, self).__init__("rouge")

  @property
  def scores_name(self):
    return {"rouge-1", "rouge-2", "rouge-l"}

  def __call__(self, ref_path, hyp_path):
    scorer = FilesRouge(metrics=list(self.scores_name))
    rouge_scores = scorer.get_scores(hyp_path, ref_path, avg=True)
    return {name:rouge_scores[name]["f"] for name in self.scores_name}


@register_scorer(name="bleu")
class BLEUScorer(Scorer):
  """Scorer using sacreBLEU."""

  def __init__(self):
    super(BLEUScorer, self).__init__("bleu")

  def __call__(self, ref_path, hyp_path):
    with tf.io.gfile.GFile(ref_path) as ref_stream, tf.io.gfile.GFile(hyp_path) as sys_stream:
      bleu = corpus_bleu(sys_stream, [ref_stream], force=True)
      return bleu.score


@register_scorer(name="wer")
class WERScorer(Scorer):
  """Scorer for WER."""

  def __init__(self):
    super(WERScorer, self).__init__("wer")

  def __call__(self, ref_path, hyp_path):
    wer_score = wer(ref_path, hyp_path)
    return wer_score

  def lower_is_better(self):
    """ Since the score shall be the lower the better """
    return True


@register_scorer(name="ter")
class TERScorer(Scorer):
  """Scorer for TER."""

  def __init__(self):
    super(TERScorer, self).__init__("ter")

  def __call__(self, ref_path, hyp_path):
    ter_score = ter(ref_path, hyp_path)
    return ter_score

  def lower_is_better(self):
    """ Since the score shall be the lower the better """
    return True


@register_scorer(name="prfmeasure", alias="prf")
class PRFScorer(Scorer):
  """Scorer for F-measure."""

  def __init__(self):
    super(PRFScorer, self).__init__("prfmeasure")

  @property
  def scores_name(self):
    return {"precision", "recall", "fmeasure"}

  def __call__(self, ref_path, hyp_path):
    precision_score, recall_score, fmeasure_score = fmeasure(ref_path, hyp_path)
    return {
        "precision": precision_score,
        "recall": recall_score,
        "fmeasure": fmeasure_score
    }


def make_scorers(names):
  """Returns a list of scorers.

  Args:
    names: A list of scorer names or a single name.

  Returns:
    A list of :class:`opennmt.evaluation.Scorer` instances.

  Raises:
    ValueError: if a scorer name is invalid.
  """
  if not isinstance(names, list):
    names = [names]
  scorers = []
  for name in names:
    scorer_class = _SCORERS_REGISTRY.get(name.lower())
    if scorer_class is None:
      raise ValueError("No scorer associated with the name: {}".format(name))
    scorers.append(scorer_class())
  return scorers
